//=============================================================================
// macro processing
//=============================================================================

#include "macros.h"

#include "directives.h"
#include "instructions.h"
#include "streams.h"
#include "assemble.h"
#include "labels.h"

const char _ENDM[] = "ENDM";
const char _MACRO[] = "MACRO";


//-----------------------------------------------------------------------------
// MACRO::add: add a macro from actual filepos
//-----------------------------------------------------------------------------
DWORD MACRO::add(char *name, char *args)
{
	DWORD ret;

	if(!name) return(RETC_PARAM_EXPECTED);
	while((*name==' ')||(*name=='\t')) name++;
	if(!*name) return(RETC_PARAM_EXPECTED);
	if(!label._validate(name)) return(RETC_UNKWN_LBL_SYNTAX);

	dbgout("MACRO::add(CHAR[" << name << "], CHAR*[" << (DWORD)args << "])\n")

	MACROLIST ML;

	//add arguments/params and check them with label::_validate
	PARAMLIST *pl;
	ML.PL = new PARAM( PARAM_GET_STRINGS );
	if((ret = ML.PL->scan( args )) != RETC_OK) return(ret);
	ret = ML.PL->ParamList->members;
	while(ret)
	{
		pl = (PARAMLIST*)ML.PL->ParamList->get(--ret);
		if(!pl) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);
		if(!label._validate(pl->str) != RETC_OK)
		{
			delete ML.PL;
			return(RETC_UNKWN_LBL_SYNTAX);
		}
	}
	ret = ML.PL->ParamList->members;
	while(ret)
	{
		pl = (PARAMLIST*)ML.PL->ParamList->get(--ret);
		if(!pl) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);
		if(instruction.if_exist(pl->str) || directive.if_exist(pl->str))
		{
			do_error(ERR_WARNING|RETC_KEY_OVERWRITE, ERR_SHOW_POS);
			break;
		}
	}

	//add the name
	ML.name = (char*)malloc(strlen(name)+1);
	if(!ML.name) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
	strcpy(ML.name, name);

	//add file info
	FILELIST *fl = stream.get();
	ML.file_no = stream.file_no;
	if(fl) ML.line_cntr = fl->line_cntr;
	else ML.line_cntr = 0;
	ML.pass = assemble.pass;

	//search if macro name already defined 4 instruction or directive
	MACROLIST *ml;
	if(ml = if_exist(name))
	{
		if(ml->pass == assemble.pass)
		{
			do_error(RETC_MACRO_OVERWRITE, ERR_SHOW_POS);
		}

		ml->pass = assemble.pass;
		ret = -1;
	}
	else ret = RETC_OK;


	//read the code for the macro
	//NOTE: if stream.in.line() is called, the parameters name,args are valid then
	char *line, *secc, *tail;

	DWORD size = 0, new_size;
	ML.code = (char *)malloc(1);
	if(!ML.code) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
	ML.code[0] = NULL;
	ML.size = 0;

	for(;;)
	{
		line = stream.in.line(STREAM_THIS_FILE);
		if(!line)
		{

			stream.file_no = ML.file_no;
			fl = stream.get();
			ret = fl->line_cntr;
			fl->line_cntr = ML.line_cntr;
			do_error(RETC_ENDM_EXPECTED, ERR_SHOW_POS);
			fl->line_cntr = ret;
			ret = RETC_OK;
			goto abort;
		}

		assemble.detect_end(line);
		line = assemble.skip_nonsens(line);

		//add 2 code
		new_size = size + strlen(line) + 1;
		ML.code = (char *)realloc(ML.code, new_size);
		if(!ML.code) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
		strcpy(&ML.code[size], line);

		secc = assemble.split_str(line);
		tail = assemble.split_str(secc);

		//test for ENDM-directive...
		if((assemble.cmp(secc, _ENDM)) && (assemble.cmp(line, ML.name)))
		{
			if((tail)&&(*tail)) do_error(ERR_WARNING|RETC_EXTRA_CHAR, ERR_SHOW_POS);
			goto code_ok;
		}

		size = new_size;
		ML.size++;  //new line completed
	}

abort:
	dbgout("free MACRO\n")
	free(ML.code);
	free(ML.name);
	delete ML.PL;
	return(ret);

code_ok:
	ML.code = (char *)realloc(ML.code, size+1);
	if(!ML.code) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
	ML.code[size] = NULL;

	if(ret != RETC_OK){ret=RETC_OK; goto abort;}

	if(instruction.if_exist(name) || directive.if_exist(name))
		do_error(ERR_WARNING|RETC_KEY_OVERWRITE, ERR_SHOW_POS);

	//add new item 2 list
	ml = (MACROLIST*)MacroList->add();
	if(!ml) do_error(ERR_FATAL|RETC_LISTALLOC, 0);

	ml->name = ML.name;
	ml->PL = ML.PL;
	ml->code = ML.code;
	ml->size = ML.size;
	ml->file_no = ML.file_no;
	ml->line_cntr = ML.line_cntr;
	ml->pass = ML.pass;

	assemble.pass_dependent=false;	//needs a new pass to check macro

	return(RETC_OK);
}


//-----------------------------------------------------------------------------
// MACRO::process: do include macro
//-----------------------------------------------------------------------------
DWORD MACRO::process(MACROLIST *ml, LIST *PL)
{
	PARAMLIST *pl, *pl_sel, *pl_new;
	char *code = ml->code;
	char *dest;
	char marks = 0;
	DWORD size = 0;
	DWORD pcnt;
	WORD len;
	BOOL possible;


	if(PL->members != ml->PL->ParamList->members) return(RETC_UNKWN_INSTR_SYNTAX);

	FILELIST *fl, *fl_src;
	if(stream.get(ml->file_no))
	{
		fl = stream.add("", FTYPE_MACRO|FTYPE_BUF);
		fl_src = stream.get(ml->file_no);
		fl->name = (char *)realloc(fl->name, strlen(fl_src->name)+strlen(ml->name)+32);
		if(!fl->name) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
		sprintf(fl->name, "%s(%i):%s", fl_src->name, ml->line_cntr, ml->name);
	}
	else
	{
		fl = stream.add(ml->name, FTYPE_MACRO|FTYPE_BUF);
	}

	fl->buf = fl->pos = (char *)malloc(MAX_LINE);
	if(!fl->pos) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
	*fl->buf = NULL;

	//do label-ovrwrt warning
	pcnt = ml->PL->ParamList->members;
	while(pcnt)
	{
		pl = (PARAMLIST*)ml->PL->ParamList->get(--pcnt);
		if(!pl) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);
		if(label.get(pl->str))
		{
			do_error(ERR_WARNING|RETC_LBL_OVERWRITE, ERR_SHOW_POS);
			break;
		}
	}

	//copy ml->code 2 a new buffer, replacing all parameters
	//do replace: param(param), 5-param+2, 5*param/2
	//dont replace: "XXXparamXXX", "param", 'param'
	//param should be sourrounded by spc/tab, coma or math-arithmetics
	//do not scan strings
	for(DWORD num_lines = ml->size; num_lines; num_lines--)
	{
		dest = fl->buf+size;
		len = 0;
		possible = true;

		while(*code)
		{
			if(possible)
			{
				for(pl_sel = NULL,pcnt = ml->PL->ParamList->members; pcnt;)
				if((pl = (PARAMLIST *)ml->PL->ParamList->get(--pcnt))&&(assemble.cmp(code, pl->str, strlen(pl->str))))
				{
					pl_sel = pl;
					pl_new = (PARAMLIST *)PL->get(pcnt);
					if(!pl_new) do_error(ERR_FATAL|RETC_LISTREQUEST, 0);
				}

				possible = false;
				if(pl_sel)
				{
					switch(*(code+strlen(pl_sel->str)))
					{
					case 0:
					case ' ': case '\t': case ',': case '@':
					case '+': case '-': case '/': case '*':
					case '%': case '(': case ')': case ':':
					case '!': case '&': case '|': case '^':
					case '>': case '<': case '.': case '~':
						if((code > ml->code)&&(*(code-1) == '<')&&(*(code+strlen(pl_sel->str)) == '>'))
						{
							code += strlen(pl_sel->str)+1;
							size += strlen(pl_new->str)-1;
							len += strlen(pl_new->str)-1;
							if(len >= MAX_LINE-1) do_error(ERR_FATAL|RETC_LINE2LONG, 0);
							strcpy(dest-1, pl_new->str);
							dest += strlen(pl_new->str)-1;
						}
						else
						{
							code += strlen(pl_sel->str);
							size += strlen(pl_new->str);
							len += strlen(pl_new->str);
							if(len >= MAX_LINE-1) do_error(ERR_FATAL|RETC_LINE2LONG, 0);
							strcpy(dest, pl_new->str);
							dest += strlen(pl_new->str);
						}
						break;
					}
				}
			}
			else
			{
				switch(*code)
				{
				case '\'':
					if(marks==1) marks = 0;
					else if(!marks) marks = 1;
					break;

				case '"':
					if(marks==2) marks = 0;
					else if(!marks) marks = 2;
					break;

				case ' ': case '\t': case ',': case '@':
				case '+': case '-': case '/': case '*':
				case '%': case '(': case ')': case ':':
				case '!': case '&': case '|': case '^':
				case '>': case '<': case '.': case '~':
					possible = true; break;
				default:
					possible = false;
				}
				if(marks) possible = false;

				*dest++ = *code++;
				if(++len >= MAX_LINE-1) do_error(ERR_FATAL|RETC_LINE2LONG, 0);
				size++;
			}
		}

		code++;
		*dest = NULL;
		size++;

		fl->pos = (char *)realloc(fl->buf, size+MAX_LINE);
		if(!fl->pos) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
		fl->buf = fl->pos;

	}

	fl->pos = (char *)realloc(fl->buf, size+1);
	if(!fl->pos) do_error(ERR_FATAL|RETC_OUTOFMEM, 0);
	fl->buf = fl->pos;
	*(fl->buf+size) = NULL;
	fl->size = ml->size;

	return(RETC_OK);
}


//-----------------------------------------------------------------------------
// MACRO::exec: execute a macro (if inside of list)
//-----------------------------------------------------------------------------
DWORD MACRO::exec(char *name, char *args)
{
	PARAM param( PARAM_GET_STRINGS );
	MACROLIST *ml;

	dbgout("MACRO::exec(CHAR[" << name << "], CHAR[" << args << "])\n")

	for(DWORD cnt = MacroList->members; cnt; )
	if(ml = (MACROLIST *)MacroList->get(--cnt))
	{
		if((ml->name)&&(assemble.cmp(ml->name, name)))
		{
			param.scan(args);

			//check if this is a directive (macro-name DIRECTIVE)
			PARAMLIST *pl = (PARAMLIST *)param.ParamList->get(0);
			if(pl)
			{
				for(int cnt=0; cnt<_D_LAST; cnt++)
				if( assemble.cmp(pl->str, directive_tbl[cnt], strlen(directive_tbl[cnt])))
				{
					switch(pl->str[ strlen(directive_tbl[cnt]) ])
					{
					case 0:	case '\t': case ' ':
					  return(RETC_UNKWN_INSTR);	//i had known it! its a directive, not a macro!
					}
				}

			}

			return( process(ml, param.ParamList) );
		}
	}

	return(RETC_UNKWN_INSTR);
}


//-----------------------------------------------------------------------------
// MACRO::if_exist: if macro exists return true
//-----------------------------------------------------------------------------
MACROLIST *MACRO::if_exist(char *name)
{
	MACROLIST *ml;
	for(DWORD cnt = MacroList->members; cnt; )
	if(ml = (MACROLIST *)MacroList->get(--cnt))
	{
		if((ml->name)&&(assemble.cmp(ml->name, name))) return(ml);
	}

	return(NULL);
}


//-----------------------------------------------------------------------------
// LABEL::constructor
//-----------------------------------------------------------------------------
MACRO::MACRO()
{
	dbgout("MACRO::MACRO()\n")

	MacroList = new LIST( sizeof(MACROLIST) );
}


//-----------------------------------------------------------------------------
// LABEL::destructor
//-----------------------------------------------------------------------------
MACRO::~MACRO()
{
	dbgout("MACRO::~MACRO()\n")

	//free MacroList (allocated bufs)
	MACROLIST *ml;
	for(DWORD cnt = MacroList->members; cnt; )
	if(ml = (MACROLIST *)MacroList->get(--cnt))
	{
		/*
		         o
		       /...\
		 \__(...  )__
		_/\ _________/

		*/

		//if(ml->fname) free(ml->fname);
		if(ml->code) free(ml->code);
		if(ml->PL) delete ml->PL;
		if(ml->name) free(ml->name);
	}

	delete MacroList;
}
